#!/usr/bin/env python3
import errno
import fcntl
import select
import shutil
import signal
import subprocess
from subprocess import PIPE
import sys
import os
import re
import string
import tempfile
import traceback
import pprint

pprinter = pprint.PrettyPrinter()


def pp(data):
	global pprinter
	hidden(pprinter.pformat(data))


mode = {}
colors = {
	"BLUE": "\033[34m",
	"GREEN": "\033[32m",
	"RED": "\033[31m",
	"YELLOW": "\033[33m",
	"BB": "\033[40m",
	"NC": "\033[0m"
}
ivpl_files = {}


def balised_output(balised, prefix, suffix, *args, **kwargs):
	end = kwargs.get("end")
	kwargs["end"] = ''
	if balised:
		print(prefix, **kwargs)
	print(*args, **kwargs)
	if balised:
		print(suffix, **kwargs)
	if end is None:
		del kwargs["end"]
	else:
		kwargs["end"] = end
	print("", **kwargs)


def colored_output(color, *args, **kwargs):
	global mode
	global colors
	balised_output(mode.get("local"), colors[color], colors["NC"], *args, **kwargs)


def comment(*args, **kwargs):
	global mode
	balised_output(not mode.get("local"), "<|--\n", "\n--|>\n", *args, **kwargs)


def title(msg):
	global mode
	if mode.get("local"):
		colored_output('BLUE', msg)
	else:
		comment("-" + msg)


def grade(value):
	hidden("Grade :=>>{:.2f}".format(value))


def error(*args, **kwargs):
	global mode
	if mode.get("local"):
		colored_output('RED', "Internal error : ", end='')
	else:
		title("Internal error")
	comment(*args, **kwargs)
	exit(1)


def warning(*args, **kwargs):
	colored_output('RED', *args, **kwargs)


def debug(*args, **kwargs):
	global mode
	if mode.get("debug"):
		colored_output('GREEN', *args, **kwargs)


def hidden(*args, **kwargs):
	colored_output('YELLOW', *args, **kwargs)


def remove(*args):
	global mode
	if not mode.get("keep_files"):
		for name in args:
			try:
				if os.path.exists(name):
					os.unlink(name)
			except Exception as e:
				error("Unlink of file resulted in " + str(e))
				return 0
	return 1


def add_values(first, second):
	for key in second.keys():
		first[key] = second[key]


def run_command(test_name, command, test, has_limits):
	try:
		os.chmod(command, 0o755)
	except OSError as e:
		error_msg = "Cannot make {} executable: {}".format(command, str(e))
		if 'compile' not in test:
			error_msg += " (and no 'compile' directive found)"
		error(error_msg)
	if 'args' not in test:
		arguments = []
	else:
		arguments = list(map(str, test['args']))

	actual_command = 'exec ' + command + ' "$@"'
	timeout = 0
	if has_limits:
		timeout = test.get('timeout', 10)
		if timeout > 0:
			arguments = [command] + arguments
			actual_command = 'exec timeout -s9 --preserve-status {} "$@"'.format(timeout)

	# We use the low level interface in order to stop collecting outputs when the limit is reached
	data = {}
	try:
		# The shell is probably impossible to avoid if we want to let a mixture of binaries,
		# well formed scripts and bare text to be accepted as executable
		# The alternative of enforcing shebang is probably a bad idea as the user might provide
		# the execute field with any interpreter launch command
		arguments = [actual_command, '--'] + arguments
		process = subprocess.Popen(arguments, stdin=PIPE, stdout=PIPE, stderr=PIPE, shell=True)
		part2stream = {
			'input': process.stdin,
			'output': process.stdout,
			'error': process.stderr
		}
		remaining = {}
		done = {}
		part2num = {}
		num2part = {}
		for part in ('input', 'output', 'error'):
			part2num[part] = part2stream[part].fileno()
			num = part2num[part]
			num2part[num] = part
			# Our I/Os are non blocking
			fcntl.fcntl(num, fcntl.F_SETFL, os.O_NONBLOCK)
			if has_limits:
				remaining[part] = test.get(part + '_limit', 10 * 1024)
			else:
				remaining[part] = -1
			done[part] = 0
			data[part] = ''
		data['input'] = str(test.get('input', ''))
		remaining['input'] = len(data['input'])
		# Debug information about what's happening
		msg = test_name + ": executing " + command
		separator = " with "
		if has_limits:
			for part in ('output', 'error'):
				msg += separator + part
				if remaining[part] >= 0:
					msg += " limited to {} bytes".format(remaining[part])
				else:
					msg += " unlimited"
				separator = ", "
			msg += " and "
			if timeout > 0:
				msg += "timeout at {}".format(timeout)
			else:
				msg += "no timeout"
		debug(msg)

		poller = select.poll()
		if remaining['input']:
			poller.register(part2num['input'], select.POLLOUT)
			set_size = 3
		else:
			part2stream['input'].close()
			set_size = 2
		poller.register(part2num['output'], select.POLLIN)
		poller.register(part2num['error'], select.POLLIN)
		while set_size > 0:
			for desc, event in poller.poll():
				part = num2part[desc]
				if remaining.get(part) >= 0:
					size = remaining[part]
				else:
					size = 1024
				try:
					if event == select.POLLOUT:
						data_bytes = data[part][done[part]:]
						number = os.write(desc, data_bytes.encode())
					else:
						data_bytes = os.read(desc, size)
						number = len(data_bytes)
						# TODO : plan a better handling of UTF-8 characters
						data[part] += data_bytes.decode(errors='replace')
					done[part] += number
					if remaining[part] != -1:
						remaining[part] -= number
					if not number or not remaining[part]:
						poller.unregister(desc)
						part2stream[part].close()
						set_size -= 1
					if not remaining[part]:
						data[part + '_bounded'] = True
				except OSError as e:
					if e.errno in [errno.EINTR, errno.EAGAIN, errno.EWOULDBLOCK]:
						debug("Access to command {} temporarily failed, {}, size {}, done {}".format(
							part, str(e), size, done[part])
						)
					else:
						debug("Error", str(e), "during access to command", part)
		if 'input' in data:
			del data['input']
		if 'input_bounded' in data:
			del data['input_bounded']
		process.wait()
		if process.returncode < 0:
			data['signal'] = -process.returncode
			data['code'] = 0
		else:
			data['signal'] = 0
			data['code'] = process.returncode
		data['failure'] = (process.returncode != 0)
	except OSError as e:
		error("Cannot exec", command, "error", str(e))
	return data


def setup_env(test_name, name, program):
	# ARGS is not necessary as it is accessible through $@
	for part in ['files']:
		if 'files' in program:
			os.environ[part.upper()] = " ".join(program['files'].keys())
	os.environ['TESTNAME'] = test_name
	os.environ['NAME'] = name


def dispatch_error(message, data, soft_fail):
	for stream in ('output', 'error'):
		data_value = data.get(stream)
		if data_value is not None and len(data_value) > 0:
			warning(data_value)
	if soft_fail:
		if len(message):
			warning(message)
	else:
		error(message)


# Runs a named command, described by a given content and associated to a test,
# returns a hash that holds the content of standard output and error as well as exit status
# The content might be :
# - scalar : a script to be executed
# - a ref to a hash : description of the executable and execution
# The test holds input and args informations
def run_special(test_name, name, content, test, has_limits):
	special = content
	default_compile = False

	if type(content) is dict:
		# Values of special are written to test to be able to overwrite intput and args in special
		# This is especially useful for a validator
		# Note here that values from the test cannot be inserted in special because, for instance,
		# the test might be compiled while special is not (this might be the case for an exercice to
		# be written in C but whose solution has been written in python
		given_test = test
		test = {}
		add_values(test, given_test)
		add_values(test, special)
	else:
		special = {'content': content}

	if 'executable' in special:
		name = special['executable']

	# sets the default compile command before populating files
	if 'compile' not in special and (
			'filename' in special and special['filename'] != name
			or ('files' in special and 'Makefile' not in special['files'])
			or 'extension' in special
	):
		default_compile = True

	# convert 'files' setting into dict format if needed
	if 'files' in special and type(special['files']) is list:
		new_files = {}
		for filename in special['files']:
			new_files[filename] = None
		special['files'] = new_files

	# generate file if required and populate the 'files' array
	if 'content' in special:
		if 'extension' in special:
			if 'filename' in special:
				error(
					"in settings for program {} in test case {}, 'extension' is incompatible with 'filename'".format(name, test_name)
				)
			special['filename'] = name + "." + special['extension']
			del special['extension']
		if 'filename' not in special:
			special['filename'] = name
		if 'files' not in special:
			special['files'] = { special['filename']: special['content'] }
		elif special['filename'] not in special['files'].keys():
			special['files'][special['filename']] = special['content']

	# write required files
	if 'files' in special:
		for filename in special['files'].keys():
			if special['files'][filename] is not None:
				ensure_write(filename, special['files'][filename])

	if default_compile:
		special['compile'] = "echo '{}: {}' | /usr/bin/env make -f - {}".format(name, " ".join(special['files'].keys()), name)
	else:
		if ('compile' not in special) and os.path.exists("Makefile") and not os.path.exists(name):
			special['compile'] = "/usr/bin/env make {}".format(name)
		else:
			if 'compile' not in special:
				special['compile'] = None

	# compile if required
	if special.get('compile'):
		if 'compiled' in special:
			tmp_name = special['compiled']
			debug("Retrieving compiled file", tmp_name, "for", name)
			shutil.copy(tmp_name, name)
		else:
			tmp, tmp_name = tempfile.mkstemp(dir=tempfile.gettempdir(), prefix=name)
			os.close(tmp)
			debug("Compiling {}, saving it to {}".format(name, tmp_name))
			setup_env(test_name, name, special)
			data = run_special(name, '.compile', special['compile'], special, False)
			if data['failure']:
				data['compilation_error'] = True
				return data
			else:
				for part in ('output', 'error'):
					if len(data[part]):
						warning("Compilation", part, ":", data['output'])
				if os.path.exists(name):
					shutil.copy(name, tmp_name)
					special['compiled'] = tmp_name

	if 'content' in special and not special.get('compile'):
		debug("Executing ", special['content'].splitlines(False)[0], "...")
	result = {}
	if 'execute' in special:
		if special['execute']:
			setup_env(test_name, name, special)
			result = run_special(test_name, '.execute', special['execute'], test, has_limits)
	else:
		result = run_command(test_name, "./{}".format(name), test, has_limits)

	# We delete the file that has been possibly created for the compilation or execution
	if 'files' in special:
		for filename in special['files'].keys():
			if special['files'][filename] is not None:
				remove(filename)
	return result


def clean_outputs(test_name, data, test):
	if 'cleanup_level' in test:
		level = test['cleanup_level']
	else:
		level = 2
	for part in ('output', 'error'):
		postprocess = None
		if 'postprocessing' in test:
			postprocess = test['postprocessing']
		if part+'_postprocessing' in test:
			postprocess = test[part+'_postprocessing']
		if postprocess is not None:
			new_data = run_special(test_name, '.' + part + '_postprocessing', postprocess, {'input': data[part]}, False)
			if len(new_data['error']):
				warning("Postprocessing produced data on stderr for {}:".format(part))
				warning(new_data['error'])
			if new_data['failure']:
				warning("*** POSSIBLE ERROR : Postprocessing of {} exited with ".format(part) + format_error(new_data))
			data[part] = new_data['output']
		if len(data[part]) and re.match(r'Grade\s*:=>>', data[part], re.IGNORECASE):
			data[part] = "Pathetic attempt to set the grade detected\n"
			data[part] += "Please stop trying to hack the system"
		cleanup(data, part, level)
		cleanup(test, part, level)


def cleanup(data, part, level):
	cleaned_data = data.get(part)
	if cleaned_data:
		cleaned_data = str(cleaned_data)
		if level > 0:
			# Removes trailing newlines
			cleaned_data = re.sub(r'\n*$', '', cleaned_data)
			if level > 0.5:
				# Removes blank lines at the beginning and the end, as well as trailing blanks
				cleaned_data = re.sub(r'^\s*\n', '', cleaned_data)
				cleaned_data = re.sub(r'\n\s*$', '', cleaned_data)
				cleaned_data = re.sub(r'[ \t]+$', '', cleaned_data, flags=re.MULTILINE)
				if level > 1:
					# Removes leading blanks
					cleaned_data = re.sub(r'^[ \t]+', '', cleaned_data, flags=re.MULTILINE)
					if level > 1.5:
						# Replaces tabs with spaces # and sequences of spaces with a single space
						cleaned_data = re.sub(r'\t', ' ', cleaned_data)
						cleaned_data = re.sub(r' +', ' ', cleaned_data)
						if level > 2:
							# Removes blank lines
							cleaned_data = re.sub(r'\n+\n', '\n', cleaned_data)
							if level > 2.5:
								# Join lines with spaces
								cleaned_data = re.sub(r'\n', ' ', cleaned_data)
		data[part] = cleaned_data


def preformat(data):
	line = str(data)
	line = re.sub(r'^(.)', r'>\1', line, flags=re.DOTALL | re.MULTILINE)
	line = re.sub(r'[^' + re.escape(string.printable) + r']', '.', line)
	if len(line) == 0 or line[-1] != '\n':
		line += '\n'
	return line


def format_error(data):
	result = ""
	if data['code']:
		result += "code " + str(data['code'])
	if data['signal']:
		if len(result):
			result += " and "
		result += "signal " + str(data['signal'])
	return result


def run_program(test_name, name, test, context, has_limits, directory):
	if directory is not None:
		original = os.getcwd()
		debug(f"Going to directory {directory} for the execution")
		os.chdir(directory)
	if 'setup' in context:
		setup = run_special(test_name, '.setup', context['setup'], context, False)
		if len(setup['output']) or len(setup['error']):
			warning("Setup output and error:")
		if setup['failure']:
			msg = "*** POSSIBLE ERROR : Setup exited with " + format_error(setup)
		else:
			msg = ""
		dispatch_error(msg, setup, 1)
	result = run_special(test_name, name, test, context, has_limits)
	if directory is not None:
		os.chdir(original)
	return result


# runs all tests for a given testcase, returns the score (percentage)
# TODO : a cleanup / restore mechanism for local files ? (using tar for instance)
def perform_tests(program, case, directories):
	global mode
	tests = case["tests"]
	score = 0
	total = 0

	if type(tests) is not dict:
		error("'tests' is not a dict for testcase " + program)
	for test_name in sorted(tests.keys()):
		if type(tests[test_name]) is not dict:
			error("Test " + test_name + " for testcase " + program + " is not a dict")
		break_loop = False
		test = {}
		add_values(test, case)
		add_values(test, tests[test_name])

		weight = test.get('weight', 1)
		total += weight
		indication = test_name
		if 'weight' in test:
			indication += " (weight {})".format(weight)
		result = None
		details = "Test {}:\n".format(test_name)

		# runs program
		data = run_program(test_name, program, test, test, True, directories.get('program'))

		if data.get('compilation_error'):
			details += "Compilation errors, compiler output :\n"
			details += preformat(data['output']) + preformat(data['error'])
			result = 0
			break_loop = True
		else:
			if test.get('compiled'):
				case['compiled'] = test['compiled']
			clean_outputs(test_name, data, test)
			if 'input' in test:
				details += "With the input:\n" + preformat(test['input'])
			if 'args' in test:
				test['args'] = list(map(str, test['args']))
				details += "With the arguments: [ {} ]\n".format(", ".join(test['args']))

			# runs a provided solution for the testcase
			# this will populate test->{output,error,code} for the following test
			if 'solution' in test:
				answer = run_program(test_name, '.solution', test['solution'], test, True, directories.get('solution'))
				# If the solution does not compile, we're in trouble
				if answer.get('compilation_error'):
					error("Solution failed to compile, stderr is\n"+answer['error'])
				# We do not test if the run fails because this might be expected by the exercise
				clean_outputs(test_name, answer, test)
				if 'match' in test:
					match = test['match']
				else:
					match = ['output']
				for part in ('output', 'error', 'code', 'signal'):
					if part in match:
						test[part] = answer[part]
					else:
						if part == 'code' or part == 'signal':
							should_display = answer[part] > 0
						else:
							should_display = len(answer[part]) > 0
						if should_display:
							debug("Solution {} :\n{}".format(part, preformat(answer[part])), end='')

			# runs a validator
			# TODO : in the current state, probably DOES NOT WORK with separate directory, investigation to be done
			if 'validator' in test:
				ensure_write(".program_name.txt", program)
				ensure_write(".arguments.txt", "\n".join(test.get('args') or []))
				for part in ('output', 'error', 'code', 'signal'):
					ensure_write(".{}.txt".format(part), data.get(part))
					if part in test:
						ensure_write(".expected_{}.txt".format(part), test[part])
				valid = run_special(test_name, '.validator', test['validator'], test, False)
				result = not valid['failure']
				for part in ('output', 'error'):
					if len(valid[part]) > 0:
						details += "Validator {} is:\n".format(part) + preformat(valid[part])
				if not result:
					details += "The validator did not validate the test ({})\n".format(format_error(valid))
				debug("Done with validator subtest with result", result)
				remove(".program_name.txt", ".arguments.txt")
				for part in ('output', 'error', 'code', 'signal'):
					remove(".{}.txt".format(part), ".expected_{}.txt".format(part))

			# compare produced output to expected one for each category
			description = {
				'output': "produced on standard output",
				'error': "produced on standard error",
				'code': "returned the code",
				'signal': "received the signal"
			}
			for part in ('output', 'error', 'code', 'signal'):
				if part in test or data[part]:
					details += "Your program {}:\n".format(description[part]) + preformat(data[part])
					if part in test:
						if data.get(part + "_bounded"):
							details += "Your program standard $part has reached the upper limit (failure)\n"
							result = False
						else:
							local_result = (data[part] == test[part])
							if not local_result:
								details += "But the expectation was:\n" + preformat(str(test[part]))
							else:
								details += "(as expected)\n"
							if result is None:
								result = True
							result = result and local_result
							debug("Done with", part, "subtest with result", local_result)
					if not result and (part == 'signal'):
						if data['signal'] == 9:
							details += "Looks like a timeout, did you write an infinite loop ?\n"
						elif data['signal'] == 13:
							details += "Looks like your program produces to much output on stdout/stderr."
							details += " This might be caused by an infinite loop.\n"
			if result is None:
				error("No method defined to check", test_name, "result !")
				result = False

		# collect result
		if result:
			debug(details, end='')
			comment("* Test", indication, "PASSED !")
			score += weight
		else:
			if not test.get('show'):
				hidden(details, end='')
				details = "(hidden)"
			comment("*", details, indication, "FAILED !")
		if break_loop:
			break
	return score / total


def read_file(name):
	try:
		with open(name, 'r') as f:
			result = f.read()
		return result
	except OSError:
		return None


def read_and_remove(name):
	global ivpl_files
	if name not in ivpl_files:
		ivpl_files[name] = read_file(name)
		remove(name)
	resultat = ivpl_files[name]
	if resultat is None:
		error("Cannot read {} !".format(name))
	return resultat


def ensure_write(name, data):
	try:
		with open(name, 'w') as f:
			if data is not None:
				f.write(str(data))
	except OSError:
		error("Cannot write {}".format(name))


def get_files(*unused, required=[], solution=[]):
	result = {}
	for file in required:
		result[file] = None
	for file in solution:
		result[file] = read_and_remove(file)
	return result


def evaluate_file(name):
	global mode
	if os.path.exists(name):
		description = read_file(name) or error("Cannot read " + name)
		try:
			exec(description, globals())
			remove(name)
		except SyntaxError as e:
			error("Evaluation of " + name + " resulted in a syntax error at line " + str(e.lineno))
		except Exception as e:
			exc_type, exc_value, exc_traceback = sys.exc_info()
			tb = traceback.extract_tb(exc_traceback)
			fs = tb.pop()
			while fs.filename != "<string>" and len(tb) > 0:
				fs = tb.pop()
			error(f'Evaluation of {name} resulted in "{str(e)}" at line {str(fs.lineno)}')


def perform_test_cases():
	global test_cases, mode
	if mode.get("local"):
		hidden("*** Running in local mode ***")
	if mode.get("debug"):
		hidden("*** Running in debug mode ***")
	debug("VPL_SUBFILES :", os.environ.get('VPL_SUBFILES'))

	# read local setup and testcases to run
	# TODO: plan a mechanism to extend the activity
	evaluate_file("vpl_evaluate.cases")

	debug("TEST CASES :\n", pprinter.pformat(test_cases))

	score = 0
	total = 0
	directories = {}
	if test_cases.get('separate_directory'):
		for name in [ 'program', 'solution' ]:
			tmpdir = tempfile.mkdtemp(prefix="iVPL_", dir=tempfile.gettempdir())
			copy_result = run_special("Global", "files_copy", f"cp -r $(ls -d * .[^.]* 2>/dev/null) '{tmpdir}'", {}, 0)
			if copy_result.get('failure'):
				dispatch_error("*** Could not copy files to temporary directory", copy_result, 0)
			os.environ[name.upper() + "_DIR"] = tmpdir
			directories[name] = tmpdir

	for name in test_cases.keys():
		if type(test_cases[name]) is not dict or 'tests' not in test_cases[name]:
			continue
		case = {}
		add_values(case, test_cases)
		add_values(case, test_cases[name])
		weight = case.get('case_weight', 1)

		case_title = "Test case " + name
		if 'case_weight' in case:
			case_title += " (weight " + str(weight) + ")"
		title(case_title)

		# Tests
		os.environ['TESTCASE'] = name
		test_score = perform_tests(name, case, directories)
		comment("=> Score for " + name + " : " + str(test_score * 100) + " / 100")
		score += test_score * weight
		total += weight

	# remove temporary directories if any
	if test_cases.get('separate_directory') and not mode.get('keep_files'):
		for key in directories.keys():
			directory=directories[key]
			debug(f"Removing temporary directory {directory}")
			remove_result = run_special("Global", 'remove', f"rm -r '{directory}'", {}, False)
			if remove_result.get('failure'):
				dispatch_error("*** Could not remove temporary directory", remove_result, False)

	min_grade = os.environ.get("VPL_GRADEMIN") or 0
	max_grade = os.environ.get("VPL_GRADEMAX") or 100
	if total == 0:
		error("No test defined !")
	grade(score * float(max_grade) / total + float(min_grade))


test_cases = {}
# We ignore sigpipes for the execution part that collects stdout and stdin
signal.signal(signal.SIGPIPE, signal.SIG_IGN)

if __name__ == '__main__':
	# *** BEGINING OF THE EVALUATION ***
	for arg in sys.argv[1:]:
		mode[arg] = True
	perform_test_cases()
